#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cerrno>

#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <string>

using namespace std;

#include "core.hh"

extern "C" { 
#include "id666.h"
#include "nez.h"
}

#include "gme/gme.h"

static FILE* resampler = NULL;
static unsigned Bufsize    = 1024;

static unsigned Rate       = 32000;
static unsigned Channels   = 2;
static unsigned Bitness    = 16;
static bool Mute = false;
static int cur_rate = 100;
static int trackno  = 0, ntracks = 0;

extern "C" int EchoEnabled; /* TODO */
extern "C" unsigned EnabledChannels; /* TODO */

static Music_Emu* emu = 0;

static void CloseResampler()
{
    if(resampler) { pclose(resampler); resampler = NULL; }
}
static void OpenResampler()
{
    int pip[2];
    socketpair(AF_UNIX, SOCK_STREAM, 0, pip);
    if(!fork())
    {
        char Buf[512];
        unsigned InputRate = 32000;
        
        std::string inputspec;
        sprintf(Buf, "r%u", InputRate);
        inputspec += Buf;


        std::string outputspec;
        if(Channels == 1) outputspec += 'm';
        if(Bitness  == 8) outputspec += 'b';
        sprintf(Buf, "r%u", Rate);
        outputspec += Buf;
        
        int p0 = dup(dup(pip[0]));
        dup2(p0, 0);
        dup2(p0, 1);
        for(int n=3; n<511; ++n) close(n);
        
        std::string filter = "linear";
        fprintf(stderr, "%s filter selected\n", filter.c_str());
        
        execlp("resample",
               "resample",
               inputspec.c_str(),
               outputspec.c_str(),
               ("-f" + filter).c_str(),
               NULL);
        
        perror("execlp(\"resample\")");
        fprintf(stderr,
            "The resample program is required for stream resampling in this\n"
            "version of spcplay. Get it at: http://bisqwit.iki.fi/source/streamixer.html\n"
        );
        _exit(-1);
    }
    close(pip[0]);
    
    resampler = fdopen(pip[1], "r+");
}


extern "C" double RunningSpeedFactor;

void SPCCore_MuteSound()        { Mute = true; }
void SPCCore_SetChannels(int n) { CloseResampler(); Channels = n; }
void SPCCore_SetBitness(int n)  { CloseResampler(); Bitness = n; }
void SPCCore_SetInterpolation(int n) { }
void SPCCore_SetEcho(int n) { EchoEnabled = n; }
void SPCCore_SetAlt(int n)  { fprintf(stderr, "alt-decoding not supported in this core\n"); }
void SPCCore_SetRate(int n) { CloseResampler(); Rate = n; }
void SPCCore_SetPlaybackSpeed(int n)
{
    cur_rate = n;
    if(n < 1) n = 1;
    RunningSpeedFactor = 100.0 / n;
}
void SPCCore_SetChannelStatus(int n, int s)
{
    if(s)
        EnabledChannels |= (1<<n);
    else
        EnabledChannels &= ~(1<<n);
}

int SPCCore_GetAlt()  { return 0; }
int SPCCore_GetEcho() { return EchoEnabled; }
int SPCCore_GetRate() { return Rate; }
int SPCCore_GetChannels() { return Channels; }
int SPCCore_GetBitness() { return Bitness; }
int SPCCore_GetChannelStatus(int n) { return EnabledChannels&(1<<n); }
int SPCCore_GetPlaybackSpeed() { return cur_rate; }
int SPCCore_GetInterpolation() { return 0; }
int SPCCore_GetNumInterpolation() { return 1; }
int SPCCore_GetNumTracks()
{
    return gme_track_count(emu);
} 
void SPCCore_SelectTrack(int trackno)
{
    fprintf(stderr, "Selecting track %d\n", trackno);
    gme_start_track(emu, trackno);
}

void SPCCore_Terminate()
{
    gme_delete(emu);
    
    CloseResampler();
#ifdef SPClog
    S9xSPClog_SaveSamples();
#endif
}

unsigned SPCCore_Init()
{
    return Bufsize;
}
int SPCCore_Load(const char* fn)
{
    char TmpFileName[] = "/tmp/gmenezXXXXXX";
    void* nezfile = 0;
    unsigned nezfilesize = NEZ_extract((char*)fn, &nezfile);
    if(!nezfilesize) return -1;
    int fd = mkstemp(TmpFileName);  
    write(fd, nezfile, nezfilesize);
    close(fd);

    int ret=0;
    //fprintf(stderr, "opening %s at rate %d\n", TmpFileName, Rate);
    const char* msg = gme_open_file(TmpFileName, &emu, Rate);
    if(msg) { ret=1; fprintf(stderr, "%s\n", msg); }
    
    if(emu) msg = gme_start_track(emu, 0);
    if(msg) { ret=1; fprintf(stderr, "%s\n", msg); }
    
    unlink(TmpFileName);
    
    if(ret==0) return 0; /* ok */
    if(ret > 1) return -1; /* file type failure */
    return -1;
}

void SPCCore_Generate(void* buffer)
{
    if(!resampler) OpenResampler();
    
    int fd = fileno(resampler);
    
    unsigned left = Bufsize;
    unsigned char *dest = (unsigned char *) buffer;
    while(left > 0)
    {
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
        
        int n = read(fd, dest, left);
        if(!n)
        {
            /* EOF */
            break;
        }
        if(n < 0)
        {
            if(errno != EAGAIN) break;
            if(errno == EINTR) continue;

            /* needs more data */
            
            fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
        
            char tmp[4096];
            const char* msg = gme_play(emu, sizeof(tmp)/2, (short*)tmp);
            if(msg)
                fprintf(stderr, "gme_play: %s\n", msg);
            write(fd, tmp, sizeof(tmp));
            //write(1, tmp, sizeof(tmp));
            
            continue;
        }
        
        left -= n;
        dest += n;
    }
}

SPCCore_ID* SPCCore_GetID(const char *fn)
{
    int fd = open(fn, O_RDONLY);
    if(fd < 0) return NULL;
    
    id666_t tmp;
    
    if(spc_read_id666(fd, &tmp) < 0)
    {
        close(fd);
        return NULL;
    }
    
    close(fd);
    
    static SPCCore_ID stat;
    memset(&stat, 0, sizeof(stat));
    
    if(tmp.song_title) strncpy(stat.songname,  tmp.song_title,  33);
    if(tmp.game_title) strncpy(stat.gametitle, tmp.game_title, 33);
    if(tmp.dumper)   strncpy(stat.dumper,    tmp.dumper,    17);
    if(tmp.comments) strncpy(stat.comments,  tmp.comments,  33);
    if(tmp.artist)   strncpy(stat.author,    tmp.artist,    33);

    stat.playtime = tmp.loop_length + tmp.intro_length;
    stat.fadetime = tmp.fade_length;
    
    if(!strcmp(tmp.emulator, "zsnes"))
       stat.emulator = 1;
    else if(!strcmp(tmp.emulator, "snes9x"))
       stat.emulator = 2;
    else
       stat.emulator = strtol(tmp.emulator, NULL, 10);
    
    return &stat;
}
